@@ -5,6 +5,7 @@ $VERBOSE = true
 $: << "../lib"
 
 require 'test/unit'
+require 'fileutils'
 require 'zip/zip'
 require 'gentestfiles'
 
@@ -94,8 +95,6 @@ class ZipEntryTest < Test::Unit::TestCase
 
     assert(entry7 != "hello")
     assert(entry7 != 12)
-
-    assert(entry7 != ZipStreamableFile.new(entry7, "aPath"))
   end
 
   def test_compare
@@ -276,12 +275,12 @@ module DecompressorTests
   end
 
   def test_readEverything
-    assert_equal(@refText, @decompressor.read)
+    assert_equal(@refText, @decompressor.sysread)
   end
     
   def test_readInChunks
     chunkSize = 5
-    while (decompressedChunk = @decompressor.read(chunkSize))
+    while (decompressedChunk = @decompressor.sysread(chunkSize))
       assert_equal(@refText.slice!(0, chunkSize), decompressedChunk)
     end
     assert_equal(0, @refText.size)
@@ -293,7 +292,7 @@ module DecompressorTests
     assert(@refLines.length > 40)
 
     
-    assert_equal(@refText[0...100], @decompressor.read(100))
+    assert_equal(@refText[0...100], @decompressor.sysread(100))
 
     assert(! @decompressor.input_finished?)
     buf = @decompressor.produce_input
@@ -403,6 +402,7 @@ class ZipInputStreamTest < Test::Unit::TestCase
   def test_new
     zis = ZipInputStream.new(TestZipFile::TEST_ZIP2.zip_name)
     assert_stream_contents(zis, TestZipFile::TEST_ZIP2)
+    assert_equal(true, zis.eof?)
     zis.close    
   end
 
@@ -410,6 +410,7 @@ class ZipInputStreamTest < Test::Unit::TestCase
     ZipInputStream.open(TestZipFile::TEST_ZIP2.zip_name) {
       |zis|
       assert_stream_contents(zis, TestZipFile::TEST_ZIP2)
+      assert_equal(true, zis.eof?)
     }
   end
 
@@ -421,19 +422,27 @@ class ZipInputStreamTest < Test::Unit::TestCase
   def test_incompleteReads
     ZipInputStream.open(TestZipFile::TEST_ZIP2.zip_name) {
       |zis|
-      entry = zis.get_next_entry
+      entry = zis.get_next_entry # longAscii.txt
+      assert_equal(false, zis.eof?)
       assert_equal(TestZipFile::TEST_ZIP2.entry_names[0], entry.name)
       assert zis.gets.length > 0
-      entry = zis.get_next_entry
+      assert_equal(false, zis.eof?)
+      entry = zis.get_next_entry # empty.txt
       assert_equal(TestZipFile::TEST_ZIP2.entry_names[1], entry.name)
       assert_equal(0, entry.size)
       assert_equal(nil, zis.gets)
-      entry = zis.get_next_entry
+      assert_equal(true, zis.eof?)
+      entry = zis.get_next_entry # empty_chmod640.txt
       assert_equal(TestZipFile::TEST_ZIP2.entry_names[2], entry.name)
-      assert zis.gets.length > 0
-      entry = zis.get_next_entry
+      assert_equal(0, entry.size)
+      assert_equal(nil, zis.gets)
+      assert_equal(true, zis.eof?)
+      entry = zis.get_next_entry # short.txt
       assert_equal(TestZipFile::TEST_ZIP2.entry_names[3], entry.name)
       assert zis.gets.length > 0
+      entry = zis.get_next_entry # longBinary.bin
+      assert_equal(TestZipFile::TEST_ZIP2.entry_names[4], entry.name)
+      assert zis.gets.length > 0
     }
   end
 
@@ -448,6 +457,7 @@ class ZipInputStreamTest < Test::Unit::TestCase
       buf << zis.read(100)
       buf << (zis.gets || "")
       buf << (zis.gets || "")
+      assert_equal(false, zis.eof?)
 
       zis.rewind
 
@@ -459,10 +469,24 @@ class ZipInputStreamTest < Test::Unit::TestCase
       assert_equal(buf, buf2)
 
       zis.rewind
+      assert_equal(false, zis.eof?)
 
       assert_entry(e.name, zis, e.name)
     }
   end
+
+  def test_mix_read_and_gets
+    ZipInputStream.open(TestZipFile::TEST_ZIP2.zip_name) {
+      |zis|
+      e = zis.get_next_entry
+      assert_equal("#!/usr/bin/env ruby", zis.gets.chomp)
+      assert_equal(false, zis.eof?)
+      assert_equal("", zis.gets.chomp)
+      assert_equal(false, zis.eof?)
+      assert_equal("$VERBOSE =", zis.read(10))
+      assert_equal(false, zis.eof?)
+    }
+  end
   
 end
 
@@ -558,7 +582,7 @@ class DeflaterTest < Test::Unit::TestCase
     File.open(fileName, "rb") {
       |file|
       inflater = Inflater.new(file)
-      txt = inflater.read
+      txt = inflater.sysread
     }
   end
 
@@ -608,6 +632,23 @@ class ZipOutputStreamTest < Test::Unit::TestCase
     end
   end
 
+  def test_put_next_entry
+    stored_text = "hello world in stored text"
+    entry_name = "file1"
+    comment = "my comment"
+    ZipOutputStream.open(TEST_ZIP.zip_name) do
+      |zos|
+      zos.put_next_entry(entry_name, comment, nil, ZipEntry::STORED)
+      zos << stored_text
+    end
+
+    assert(File.read(TEST_ZIP.zip_name).grep(stored_text))
+    ZipFile.open(TEST_ZIP.zip_name) do
+      |zf|
+      assert_equal(stored_text, zf.read(entry_name))
+    end
+  end
+
   def assert_i_o_error_in_closed_stream
     assert_raise(IOError) {
       zos = ZipOutputStream.new("test_putOnClosedStream.zip")
@@ -813,6 +854,28 @@ class ZipEntrySetTest < Test::Unit::TestCase
     assert_equal(entries[3], entrySet.parent(entries[4]))
     assert_equal(entries[3], entrySet.parent(entries[5]))
   end
+
+  def test_glob
+    res = @zipEntrySet.glob('name[2-4]')
+    assert_equal(3, res.size)
+    assert_equal(ZIP_ENTRIES[1,3], res)
+  end
+
+  def test_glob2
+    entries = [ 
+      ZipEntry.new("zf.zip", "a/"),
+      ZipEntry.new("zf.zip", "a/b/b1"),
+      ZipEntry.new("zf.zip", "a/b/c/"),
+      ZipEntry.new("zf.zip", "a/b/c/c1")
+    ]
+    entrySet = ZipEntrySet.new(entries)
+
+    assert_equal(entries[0,1], entrySet.glob("*"))
+#    assert_equal(entries[FIXME], entrySet.glob("**"))
+#    res = entrySet.glob('a*')
+#    assert_equal(entries.size, res.size)
+#    assert_equal(entrySet.map { |e| e.name }, res.map { |e| e.name })
+  end
 end
 
 
@@ -967,11 +1030,11 @@ module CommonZipFileFixture
   EMPTY_FILENAME = "emptyZipFile.zip"
 
   TEST_ZIP = TestZipFile::TEST_ZIP2.clone
-  TEST_ZIP.zip_name = "4entry_copy.zip"
+  TEST_ZIP.zip_name = "5entry_copy.zip"
 
   def setup
     File.delete(EMPTY_FILENAME) if File.exists?(EMPTY_FILENAME)
-    File.copy(TestZipFile::TEST_ZIP2.zip_name, TEST_ZIP.zip_name)
+    FileUtils.cp(TestZipFile::TEST_ZIP2.zip_name, TEST_ZIP.zip_name)
   end
 end
 
@@ -1011,13 +1074,18 @@ class ZipFileTest < Test::Unit::TestCase
       assert_equal(entryCount+1, zf.size)
       assert_equal("Putting stuff in data/generated/empty.txt", zf.read("data/generated/empty.txt")) 
 
+      zf.get_output_stream('entry.bin') {
+	|os|
+	os.write(File.open('data/generated/5entry.zip', 'rb').read)
+      }
     }
     
     ZipFile.open(TEST_ZIP.zip_name) {
       |zf|
-      assert_equal(entryCount+1, zf.size)
+      assert_equal(entryCount+2, zf.size)
       assert_equal("Putting stuff in newEntry.txt", zf.read("newEntry.txt")) 
       assert_equal("Putting stuff in data/generated/empty.txt", zf.read("data/generated/empty.txt")) 
+      assert_equal(File.open('data/generated/5entry.zip', 'rb').read, zf.read("entry.bin")) 
     }
   end
 
@@ -1076,7 +1144,7 @@ class ZipFileTest < Test::Unit::TestCase
   def test_remove
     entryToRemove, *remainingEntries = TEST_ZIP.entry_names
 
-    File.copy(TestZipFile::TEST_ZIP2.zip_name, TEST_ZIP.zip_name)
+    FileUtils.cp(TestZipFile::TEST_ZIP2.zip_name, TEST_ZIP.zip_name)
 
     zf = ZipFile.new(TEST_ZIP.zip_name)
     assert(zf.entries.map { |e| e.name }.include?(entryToRemove))
@@ -1091,24 +1159,27 @@ class ZipFileTest < Test::Unit::TestCase
     zfRead.close
   end
 
-
   def test_rename
     entryToRename, *remainingEntries = TEST_ZIP.entry_names
-    
+
     zf = ZipFile.new(TEST_ZIP.zip_name)
     assert(zf.entries.map { |e| e.name }.include?(entryToRename))
-    
-    newName = "changed name"
+
+    contents = zf.read(entryToRename)
+    newName = "changed entry name"
     assert(! zf.entries.map { |e| e.name }.include?(newName))
 
     zf.rename(entryToRename, newName)
     assert(zf.entries.map { |e| e.name }.include?(newName))
 
+    assert_equal(contents, zf.read(newName))
+
     zf.close
 
     zfRead = ZipFile.new(TEST_ZIP.zip_name)
     assert(zfRead.entries.map { |e| e.name }.include?(newName))
-    zfRead.close
+    assert_equal(contents, zf.read(newName))
+    zfRead.close    
   end
 
   def test_renameToExistingEntry
@@ -1221,12 +1292,12 @@ class ZipFileTest < Test::Unit::TestCase
   # can delete the file you used to add the entry to the zip file
   # with
   def test_commitUseZipEntry
-    File.copy(TestFiles::RANDOM_ASCII_FILE1, "okToDelete.txt")
+    FileUtils.cp(TestFiles::RANDOM_ASCII_FILE1, "okToDelete.txt")
     zf = ZipFile.open(TEST_ZIP.zip_name)
     zf.add("okToDelete.txt", "okToDelete.txt")
     assert_contains(zf, "okToDelete.txt")
     zf.commit
-    File.move("okToDelete.txt", "okToDeleteMoved.txt")
+    File.rename("okToDelete.txt", "okToDeleteMoved.txt")
     assert_contains(zf, "okToDelete.txt", "okToDeleteMoved.txt")
   end
 
@@ -1349,6 +1420,17 @@ class ZipFileExtractTest < Test::Unit::TestCase
       assert(File.exists?(EXTRACTED_FILENAME))
       AssertEntry::assert_contents(EXTRACTED_FILENAME, 
 				  zf.get_input_stream(ENTRY_TO_EXTRACT) { |is| is.read })
+
+
+      File::unlink(EXTRACTED_FILENAME)
+
+      entry = zf.get_entry(ENTRY_TO_EXTRACT)
+      entry.extract(EXTRACTED_FILENAME)
+
+      assert(File.exists?(EXTRACTED_FILENAME))
+      AssertEntry::assert_contents(EXTRACTED_FILENAME, 
+				  entry.get_input_stream() { |is| is.read })
+
     }
   end
 
@@ -1465,25 +1547,6 @@ class ZipFileExtractDirectoryTest < Test::Unit::TestCase
   end
 end
 
-class ZipStreamableFileTest < Test::Unit::TestCase
-  def test_equality
-    zipEntry1 = ZipEntry.new("zf.zip", "entryname1", "comment")
-    zipEntry2 = ZipEntry.new("zf.zip", "entryname2", "comment")
-
-    zipStreamableFile1 = ZipStreamableFile.new(zipEntry1, "path")
-    zipStreamableFile2 = ZipStreamableFile.new(zipEntry1, "path")
-    zipStreamableFile3 = ZipStreamableFile.new(zipEntry1, "anotherPath")
-    zipStreamableFile4 = ZipStreamableFile.new(zipEntry2, "path")
-    
-    assert_equal(zipStreamableFile1, zipStreamableFile1)
-    assert_equal(zipStreamableFile1, zipStreamableFile2)
-    assert(zipStreamableFile1 != zipStreamableFile3)
-    assert(zipStreamableFile1 != zipStreamableFile4)
-    assert(zipStreamableFile1 != zipEntry1)
-    assert(zipStreamableFile1 != "hej")
-  end
-end
-
 class ZipExtraFieldTest < Test::Unit::TestCase
   def test_new
     extra_pure    = ZipExtraField.new("")
@@ -1552,6 +1615,6 @@ class ZipExtraFieldTest < Test::Unit::TestCase
 
 end
 
-# Copyright (C) 2002, 2003 Thomas Sondergaard
+# Copyright (C) 2002-2005 Thomas Sondergaard
 # rubyzip is free software; you can redistribute it and/or
 # modify it under the terms of the ruby license.
